# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
from pytest import approx

import sisl
from sisl.io.siesta.fdf import *
from sisl.io.siesta.stdout import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_mgco3_md_out(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    out = stdoutSileSiesta(f)

    geom0 = out.read_geometry()
    geom = out.read_geometry[-1]()
    geom1 = out.read_data(geometry=True, slice=-1)

    # assert it works correct
    assert isinstance(geom0, sisl.Geometry)
    assert isinstance(geom, sisl.Geometry)
    assert isinstance(geom1, sisl.Geometry)
    # assert first and last are not the same
    assert not np.allclose(geom0.xyz, geom.xyz)
    assert not np.allclose(geom0.xyz, geom1.xyz)

    # try and read all outputs (including the final section)
    nOutputs = 6
    assert len(out.read_force[:](skip_final=False)) == nOutputs
    assert len(out.read_stress[:](skip_final=False)) == nOutputs

    # there are 5 dynamics outputs in this output file.
    nOutputs = 5
    assert len(out.read_geometry[:]()) == nOutputs
    assert len(out.read_force[:]()) == nOutputs
    assert len(out.read_stress[:]()) == nOutputs

    f0 = out.read_force()
    f = out.read_force[-1]()
    f1 = out.read_data(force=True, slice=-1)
    assert not np.allclose(f0, f)
    assert np.allclose(f1, f)

    # Check that we can read the different types of forces
    nAtoms = 10
    atomicF = out.read_force[:]()
    totalF = out.read_force[:](total=True)
    maxF = out.read_force[:](max=True)
    assert atomicF.shape == (nOutputs, nAtoms, 3)
    assert totalF.shape == (nOutputs, 3)
    assert maxF.shape == (nOutputs,)
    totalF, maxF = out.read_force(total=True, max=True)
    assert totalF.shape == (3,)
    assert maxF.shape == ()

    s0 = out.read_stress()
    s = out.read_stress[-1]()
    assert not np.allclose(s0, s)

    sstatic = out.read_stress[:]("static")
    stotal = out.read_stress[:]("total")

    for S, T in zip(sstatic, stotal):
        assert not np.allclose(S, T)


def test_mgco3_md_out_scf(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    out = stdoutSileSiesta(f)

    # Ensure SCF reads are consistent
    scf_last = out.read_scf[:]()
    scf_last2, props = out.read_scf[:](ret_header=True)
    scf = out.read_scf[-1]()
    scf_props = out.read_scf[-1](ret_header=True)
    assert scf_props[1] == props
    assert np.allclose(scf, scf_props[0])
    assert np.allclose(scf_last[-1], scf)
    for i in range(len(scf_last)):
        scf = out.read_scf[i]()
        assert np.allclose(scf_last[i], scf)
        assert np.allclose(scf_last[i], scf_last2[i])

    scf_all = out.read_scf[-1](iscf=None)
    scf = out.read_scf[-1]()
    assert np.allclose(scf_all[-1], scf)
    for i in range(len(scf_all)):
        scf = out.read_scf[-1](iscf=i + 1)
        assert np.allclose(scf_all[i], scf)


def test_mgco3_md_out_data(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    out = stdoutSileSiesta(f)

    f0, g0 = out.read_data(force=True, geometry=True)
    g1, f1, e = out.read_data(geometry=True, force=True, energy=True)

    assert np.allclose(f0, f1)
    assert g0 == g1
    assert isinstance(e, sisl.utils.PropertyDict)
    assert e.fermi == approx(-3.420926)
    assert e.xc == approx(-1218.701737)
    assert e["kinetic"] == approx(3955.286834)


def test_mgco3_md_out_info(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    out = stdoutSileSiesta(f)
    assert out.info.completed
    assert out.info.spin.is_unpolarized
    geom = out.read_geometry()
    assert out.info.na == geom.na
    assert out.info.no == geom.no


def test_mgco3_md_out_dataframe(sisl_files):
    pytest.importorskip("pandas", reason="pandas not available")
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    out = stdoutSileSiesta(f)

    data = out.read_scf[:]()
    df = out.read_scf[:](as_dataframe=True)
    # this will read all MD-steps and only latest iscf
    assert len(data) == len(df)
    assert df.index.names == ["imd"]

    df = out.read_scf[:](iscf=None, as_dataframe=True)
    assert df.index.names == ["imd", "iscf"]
    df = out.read_scf(iscf=None, as_dataframe=True)
    assert df.index.names == ["iscf"]


def test_mgco3_md_out_energy(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")
    energy = stdoutSileSiesta(f).read_energy()
    assert isinstance(energy, sisl.utils.PropertyDict)
    assert hasattr(energy, "basis")
    basis = energy.basis
    assert hasattr(basis, "enthalpy")


def test_mgco3_md_pao_basis(sisl_files):
    f = sisl_files("siesta", "MgCO3_md", "RUN.out")

    block = """
Mg                    3                    # Species label, number of l-shells
 n=2   0   1                         # n, l, Nzeta
   2.346
   1.000
 n=3   0   1                         # n, l, Nzeta
   7.081
   1.000
 n=2   1   1                         # n, l, Nzeta
   2.614
   1.000
C                     2                    # Species label, number of l-shells
 n=2   0   1                         # n, l, Nzeta
   4.511
   1.000
 n=2   1   1                         # n, l, Nzeta
   5.490
   1.000
O                     2                    # Species label, number of l-shells
 n=2   0   1                         # n, l, Nzeta
   3.561
   1.000
 n=2   1   1                         # n, l, Nzeta
   4.343
   1.000
    """

    atom_orbs = fdfSileSiesta._parse_pao_basis(block)
    assert len(atom_orbs) == 3
    assert len(atom_orbs["Mg"]) == 5
    assert len(atom_orbs["C"]) == 4
    assert len(atom_orbs["O"]) == 4

    atoms = stdoutSileSiesta(f).read_basis()
    for atom in atoms:
        assert atom.orbitals == atom_orbs[atom.tag]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.shape._cylinder import *
from sisl.utils.mathematics import fnorm

pytestmark = pytest.mark.shape


@pytest.mark.filterwarnings("ignore", message="*orthogonalizes the vectors")
def test_create_ellipticalcylinder():
    el = EllipticalCylinder(1.0, 1.0)
    el = EllipticalCylinder([1.0, 1.0], 1.0)
    v0 = [1.0, 0.2, 1.0]
    v1 = [1.0, -0.2, 1.0]
    el = EllipticalCylinder([v0, v1], 1.0)
    v0 = el.radial_vector[0]
    v1 = el.radial_vector[1]
    v0 /= fnorm(v0)
    v1 /= fnorm(v1)
    el = EllipticalCylinder([v0, v1], 1.0)
    e2 = el.scale(1.1)
    assert np.allclose(el.radius + 0.1, e2.radius)
    assert np.allclose(el.height + 0.1, e2.height)
    e2 = el.scale([1.1] * 3)
    assert np.allclose(el.radius + 0.1, e2.radius)
    assert np.allclose(el.height + 0.1, e2.height)
    e2 = el.scale([1.1, 2.1, 3.1])
    assert np.allclose(el.radius + [0.1, 1.1], e2.radius)
    assert np.allclose(el.height + 2.1, e2.height)
    str(el)
    e2.volume


def test_ellipticalcylinder_within():
    el = EllipticalCylinder(1.0, 1.0)
    # center of cylinder
    assert el.within_index([0, 0, 0])[0] == 0
    # should not be in a circle
    assert el.within_index([0.2, 0.2, 0.9])[0] == 0


def test_tosphere():
    el = EllipticalCylinder([1.0, 1.0], 1.0)
    el.to.Sphere()


def test_tocuboid():
    el = EllipticalCylinder([1.0, 1.0], 1.0)
    el.to.Cuboid()


def test_translate():
    el = EllipticalCylinder([1.0, 1.0], 1.0)
    el2 = el.translate([0, 1, 2])
    assert not np.allclose(el.center, el2.center)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import operator as ops

import pytest

from sisl import _array as ar
from sisl._core.oplist import oplist

pytestmark = pytest.mark.oplist


def test_oplist_creation():
    l = oplist(range(10))
    assert len(l) == 10
    l = oplist([1, 2])
    assert len(l) == 2
    l = oplist((1, 2))
    assert len(l) == 2
    assert l[0] == 1
    assert l[1] == 2


@pytest.mark.parametrize(
    "op", [ops.add, ops.floordiv, ops.sub, ops.mul, ops.truediv, ops.pow]
)
@pytest.mark.parametrize("key1", [1, 2])
@pytest.mark.parametrize("key2", [1, 2])
def test_oplist_math(op, key1, key2):
    d = {
        1: oplist([ar.aranged(1, 10), ar.aranged(1, 10)]),
        2: 2,
    }

    l1 = d[key1]
    l2 = d[key2]
    op(l1, l2)


@pytest.mark.parametrize("op", [ops.abs, ops.neg, ops.pos])
def test_oplist_single(op):
    d = oplist([ar.aranged(1, 10), ar.aranged(1, 10)])
    op(d)


@pytest.mark.parametrize(
    "op", [ops.iadd, ops.ifloordiv, ops.isub, ops.imul, ops.itruediv, ops.ipow]
)
@pytest.mark.parametrize("key", [1, 2])
def test_oplist_imath(op, key):
    d = {
        1: oplist([ar.aranged(1, 10), ar.aranged(1, 10)]),
        2: 2,
    }

    l2 = d[key]
    try:
        l1 = oplist(data * 2 for data in l2)
    except Exception:
        l1 = oplist([ar.aranged(1, 10), ar.aranged(2, 3)])
    op(l1, l2)


@pytest.mark.parametrize("op", [ops.add, ops.sub, ops.mul, ops.truediv, ops.pow])
def test_oplist_fail_math(op):
    l1 = oplist([1, 2])
    l2 = [1, 2, 3]
    with pytest.raises(ValueError):
        op(l1, l2)


@pytest.mark.parametrize("op", [ops.sub, ops.mul, ops.truediv, ops.pow])
def test_oplist_fail_rmath(op):
    l1 = oplist([1, 2])
    l2 = [1, 2, 3]
    with pytest.raises(ValueError):
        op(l2, l1)


@pytest.mark.parametrize("op", [ops.iadd, ops.isub, ops.imul, ops.itruediv, ops.ipow])
def test_oplist_fail_imath(op):
    l1 = oplist([1, 2])
    l2 = [1, 2, 3]
    with pytest.raises(ValueError):
        op(l1, l2)


def test_oplist_deco():
    @oplist.decorate
    def my_func():
        return 1

    a = my_func()
    assert isinstance(a, oplist)
    assert a[0] == 1

    @oplist.decorate
    def my_func():
        return [2, 3]

    a = my_func()
    assert isinstance(a, oplist)
    assert len(a) == 2
    assert a[0] == 2

    @oplist.decorate
    def my_func():
        return oplist([1, 2])

    a = my_func()
    assert isinstance(a, oplist)
    assert len(a) == 2
    assert a[0] == 1
    assert a[1] == 2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.viz.processors.axes import (
    axes_cross_product,
    axis_direction,
    get_ax_title,
    sanitize_axes,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


def test_sanitize_axes():
    assert sanitize_axes(["x", "y", "z"]) == ["x", "y", "z"]
    assert sanitize_axes("xyz") == ["x", "y", "z"]
    assert sanitize_axes("abc") == ["a", "b", "c"]
    assert sanitize_axes([0, 1, 2]) == ["a", "b", "c"]
    assert sanitize_axes("-xy") == ["-x", "y"]
    assert sanitize_axes("x-y") == ["x", "-y"]
    assert sanitize_axes("-x-y") == ["-x", "-y"]
    assert sanitize_axes("a-b") == ["a", "-b"]

    axes = sanitize_axes([[0, 1, 2]])
    assert isinstance(axes[0], np.ndarray)
    assert axes[0].shape == (3,)
    assert np.all(axes[0] == [0, 1, 2])

    with pytest.raises(ValueError):
        sanitize_axes([None])


def test_axis_direction():
    assert np.allclose(axis_direction("x"), [1, 0, 0])
    assert np.allclose(axis_direction("y"), [0, 1, 0])
    assert np.allclose(axis_direction("z"), [0, 0, 1])

    assert np.allclose(axis_direction("-x"), [-1, 0, 0])
    assert np.allclose(axis_direction("-y"), [0, -1, 0])
    assert np.allclose(axis_direction("-z"), [0, 0, -1])

    assert np.allclose(axis_direction([1, 0, 0]), [1, 0, 0])

    cell = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])

    assert np.allclose(axis_direction("a", cell), [0, 0, 1])
    assert np.allclose(axis_direction("b", cell), [1, 0, 0])
    assert np.allclose(axis_direction("c", cell), [0, 1, 0])

    assert np.allclose(axis_direction("-a", cell), [0, 0, -1])
    assert np.allclose(axis_direction("-b", cell), [-1, 0, 0])
    assert np.allclose(axis_direction("-c", cell), [0, -1, 0])


def test_axes_cross_product():
    assert np.allclose(axes_cross_product("x", "y"), [0, 0, 1])
    assert np.allclose(axes_cross_product("y", "x"), [0, 0, -1])
    assert np.allclose(axes_cross_product("-x", "y"), [0, 0, -1])

    assert np.allclose(axes_cross_product([1, 0, 0], [0, 1, 0]), [0, 0, 1])
    assert np.allclose(axes_cross_product([0, 1, 0], [1, 0, 0]), [0, 0, -1])

    cell = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])

    assert np.allclose(axes_cross_product("b", "c", cell), [0, 0, 1])
    assert np.allclose(axes_cross_product("c", "b", cell), [0, 0, -1])
    assert np.allclose(axes_cross_product("-b", "c", cell), [0, 0, -1])


def test_axis_title():
    assert get_ax_title("title") == "title"

    assert get_ax_title("x") == "X axis [Ang]"
    assert get_ax_title("y") == "Y axis [Ang]"
    assert get_ax_title("-z") == "-Z axis [Ang]"

    assert get_ax_title("a") == "A lattice vector"
    assert get_ax_title("b") == "B lattice vector"
    assert get_ax_title("-c") == "-C lattice vector"

    assert get_ax_title(None) == ""

    assert get_ax_title(np.array([1, 2, 3])) == "[1 2 3]"

    def some_axis():
        pass

    assert get_ax_title(some_axis) == "some_axis"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl
from sisl.viz.data import BandsData

pytestmark = [pytest.mark.viz, pytest.mark.data]


@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def test_bands_from_sisl_H(spin):
    gr = sisl.geom.graphene()
    H = sisl.Hamiltonian(gr)
    H.construct([(0.1, 1.44), (0, -2.7)])

    n_spin, H = {
        "unpolarized": (1, H),
        "polarized": (2, H.transform(spin=sisl.Spin.POLARIZED)),
        "noncolinear": (4, H.transform(spin=sisl.Spin.NONCOLINEAR)),
        "spinorbit": (4, H.transform(spin=sisl.Spin.SPINORBIT)),
    }[spin]

    bz = sisl.BandStructure(
        H, [[0, 0, 0], [2 / 3, 1 / 3, 0], [1 / 2, 0, 0]], 6, ["Gamma", "M", "K"]
    )

    data = BandsData.new(bz)

    data.sanity_check(
        n_spin=n_spin,
        nk=6,
        nbands=2,
        klabels=["Gamma", "M", "K"],
        kvals=[0.0, 1.70309799, 2.55464699],
    )


def test_bands_from_siesta_bands(sisl_files):
    n_spin = 1
    file = sisl_files("siesta", "SrTiO3", "unpolarized", "SrTiO3.bands")

    data = BandsData.new(file)

    data.sanity_check(
        n_spin=n_spin,
        nk=150,
        nbands=72,
        klabels=("Gamma", "X", "M", "Gamma", "R", "X"),
        kvals=[0.0, 0.429132, 0.858265, 1.465149, 2.208428, 2.815313],
    )


@pytest.mark.parametrize("spin", ["noncolinear"])
def test_bands_from_siesta_wfsx(spin, sisl_files):
    n_spin = 4
    dirs = ("siesta", "Bi2Se3_3layer")

    # From a siesta .WFSX file
    # Since there is no hamiltonian for bi2se3_3ql.fdf, we create a dummy one
    wfsx = sisl.get_sile(sisl_files(*dirs, "Bi2Se3.bands.WFSX"))

    fdf = sisl_files(*dirs, "Bi2Se3.fdf")

    data = BandsData.new(wfsx, fdf=fdf)

    data.sanity_check(n_spin=n_spin, nk=16, nbands=4)


@pytest.mark.parametrize(
    "spin", ["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def test_toy_example(spin):
    nk = 15
    n_states = 28

    data = BandsData.toy_example(spin=spin, nk=nk, n_states=n_states)

    n_spin = {"unpolarized": 1, "polarized": 2, "noncolinear": 4, "spinorbit": 4}[spin]

    data.sanity_check(
        n_spin=n_spin, nk=nk, nbands=n_states, klabels=["Gamma", "X"], kvals=[0, 1]
    )

    if n_spin == 4:
        assert "spin_moments" in data.data_vars



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

import sisl
from sisl import AtomicOrbital, Geometry
from sisl.messages import SislError
from sisl.viz.data import PDOSData
from sisl.viz.processors.orbital import (
    OrbitalQueriesManager,
    atom_data_from_orbital_data,
    reduce_orbital_data,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


@pytest.fixture(scope="module")
def geometry():
    orbs = [
        AtomicOrbital("2sZ1"),
        AtomicOrbital("2sZ2"),
        AtomicOrbital("2pxZ1"),
        AtomicOrbital("2pyZ1"),
        AtomicOrbital("2pzZ1"),
        AtomicOrbital("2pxZ2"),
        AtomicOrbital("2pyZ2"),
        AtomicOrbital("2pzZ2"),
    ]

    atoms = [
        sisl.Atom(5, orbs),
        sisl.Atom(7, orbs),
    ]
    return sisl.geom.graphene(atoms=atoms)


@pytest.fixture(
    scope="module", params=["unpolarized", "polarized", "noncolinear", "spinorbit"]
)
def spin(request):
    return sisl.Spin(request.param)


@pytest.fixture(scope="module")
def orb_manager(geometry, spin):
    return OrbitalQueriesManager(geometry, spin=spin)


def test_get_orbitals(orb_manager, geometry: Geometry):
    orbs = orb_manager.get_orbitals({"atoms": [0]})
    assert len(orbs) == geometry.atoms.atom[0].no
    assert np.all(orbs == np.arange(geometry.atoms.atom[0].no))

    orbs = orb_manager.get_orbitals({"orbitals": [0, 1]})
    assert len(orbs) == 2
    assert np.all(orbs == np.array([0, 1]))


def test_get_atoms(orb_manager, geometry: Geometry):
    ats = orb_manager.get_atoms({"atoms": [0]})
    assert len(ats) == 1
    assert ats[0] == 0

    if geometry.orbitals[0] > 1:
        at_orbitals = [0, 1]
    else:
        at_orbitals = [0]

    ats = orb_manager.get_atoms({"orbitals": at_orbitals})
    assert len(ats) == 1
    assert np.all(ats == np.array([0]))


def test_split(orb_manager, geometry: Geometry):
    # Check that it can split over species
    queries = orb_manager.generate_queries(split="species")

    assert len(queries) == geometry.atoms.nspecies

    atom_tags = [atom.tag for atom in geometry.atoms.atom]

    for query in queries:
        assert isinstance(query, dict), f"Query is not a dict: {query}"

        assert "species" in query, f"Query does not have species: {query}"
        assert isinstance(query["species"], list)
        assert len(query["species"]) == 1
        assert query["species"][0] in atom_tags

    # Check that it can split over atoms
    queries = orb_manager.generate_queries(split="atoms")

    assert len(queries) == geometry.na

    for i_atom in range(geometry.na):
        query = queries[i_atom]

        assert isinstance(query, dict), f"Query is not a dict: {query}"

        assert "atoms" in query, f"Query does not have atoms: {query}"
        assert isinstance(query["atoms"], list)
        assert len(query["atoms"]) == 1
        assert query["atoms"][0] == i_atom

    # Check that it can split over n
    queries = orb_manager.generate_queries(split="l")

    assert len(queries) != 0

    for query in queries:
        assert isinstance(query, dict), f"Query is not a dict: {query}"

        assert "l" in query, f"Query does not have l: {query}"
        assert isinstance(query["l"], list)
        assert len(query["l"]) == 1
        assert isinstance(query["l"][0], int)


def test_double_split(orb_manager):
    # Check that it can split over two things at the same time
    queries = orb_manager.generate_queries(split="l+m")

    assert len(queries) != 0

    for query in queries:
        assert isinstance(query, dict), f"Query is not a dict: {query}"

        assert "l" in query, f"Query does not have l: {query}"
        assert isinstance(query["l"], list)
        assert len(query["l"]) == 1
        assert isinstance(query["l"][0], int)

        assert "l" in query, f"Query does not have l: {query}"
        assert isinstance(query["m"], list)
        assert len(query["m"]) == 1
        assert isinstance(query["m"][0], int)

        assert abs(query["m"][0]) <= query["l"][0]


def test_split_only(orb_manager, geometry):
    queries = orb_manager.generate_queries(
        split="species", only=[geometry.atoms.atom[0].tag]
    )

    assert len(queries) == 1
    assert queries[0]["species"] == [geometry.atoms.atom[0].tag]


def test_split_exclude(orb_manager, geometry):
    queries = orb_manager.generate_queries(
        split="species", exclude=[geometry.atoms.atom[0].tag]
    )

    assert len(queries) == geometry.atoms.nspecies - 1
    assert geometry.atoms.atom[0].tag not in [query["species"][0] for query in queries]


def test_constrained_split(orb_manager, geometry):
    queries = orb_manager.generate_queries(split="species", atoms=[0])

    assert len(queries) == 1
    assert queries[0]["species"] == [geometry.atoms.atom[0].tag]


def test_split_name(orb_manager, geometry):
    queries = orb_manager.generate_queries(split="species", name="Tag: $species")

    assert len(queries) == geometry.atoms.nspecies

    for query in queries:
        assert "name" in query, f"Query does not have name: {query}"
        assert query["name"] == f"Tag: {query['species'][0]}"


def test_sanitize_query(orb_manager, geometry):
    san_query = orb_manager.sanitize_query({"atoms": [0]})

    atom_orbitals = geometry.atoms.atom[0].orbitals

    assert len(san_query["orbitals"]) == len(atom_orbitals)
    assert np.all(san_query["orbitals"] == np.arange(len(atom_orbitals)))


def test_reduce_orbital_data(geometry, spin):
    data = PDOSData.toy_example(geometry=geometry, spin=spin)._data

    reduced = reduce_orbital_data(data, [{"name": "all"}])

    assert isinstance(reduced, xr.DataArray)

    for dim in data.dims:
        if dim == "orb":
            assert dim not in reduced.dims
        else:
            assert dim in reduced.dims
            assert len(data[dim]) == len(reduced[dim])

    assert "group" in reduced.dims
    assert len(reduced.group) == 1
    assert reduced.group[0] == "all"
    assert np.allclose(reduced.sel(group="all").values, data.sum("orb").values)

    data_no_geometry = data.copy()
    data_no_geometry.attrs.pop("geometry")

    with pytest.raises(SislError):
        reduced = reduce_orbital_data(data_no_geometry, [{"name": "all"}])


def test_reduce_orbital_data_spin(geometry, spin):
    data = PDOSData.toy_example(geometry=geometry, spin=spin)._data

    if spin.is_polarized:
        sel_total = reduce_orbital_data(data, [{"name": "all", "spin": "total"}])
        red_total = reduce_orbital_data(data, [{"name": "all"}], spin_reduce=np.sum)

        assert np.allclose(sel_total.values, red_total.values)


def test_reduce_orbital_data_no_spin(geometry, spin):
    """Reduce orbital data without reducing the spin dimension"""
    data = PDOSData.toy_example(geometry=geometry, spin=spin)._data

    reduced = reduce_orbital_data(
        data, [{"name": "all", "spin": "total"}], spin_reduce=False
    )

    assert np.all(data.spin == reduced.spin)


def test_atom_data_from_orbital_data(geometry: Geometry, spin):
    data = PDOSData.toy_example(geometry=geometry, spin=spin)._data

    atom_data = atom_data_from_orbital_data(data, geometry)

    assert isinstance(atom_data, xr.DataArray)

    for dim in data.dims:
        if dim == "orb":
            assert dim not in atom_data.dims
        else:
            assert dim in atom_data.dims
            assert len(data[dim]) == len(atom_data[dim])

    assert "atom" in atom_data.dims
    assert len(atom_data.atom) == geometry.na
    assert np.all(atom_data.atom == np.arange(geometry.na))

    atom_values = []
    firsto = geometry.firsto
    lasto = geometry.lasto
    for i in range(geometry.na):
        atom_values.append(data.sel(orb=slice(firsto[i], lasto[i])).sum("orb").values)

    assert np.allclose(atom_data.values, np.array(atom_values))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

from sisl.io.orca.txt import *

pytestmark = [pytest.mark.io, pytest.mark.orca]


def test_tags(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    assert out.info.na == 2


def test_read_electrons(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    N = out.read_electrons[:]()
    assert N[0, 0] == pytest.approx(7.9999985377)
    assert N[0, 1] == pytest.approx(6.9999989872)
    N = out.read_electrons[-1]()
    assert N[0] == pytest.approx(7.9999985377)
    assert N[1] == pytest.approx(6.9999989872)


def test_read_energy(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    E = out.read_energy[:]()
    assert len(E) == 2
    E = out.read_energy[-1]()
    assert pytest.approx(E.total) == pytest.approx(-3532.4797529097723)


def test_read_energy_vdw(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule_property.txt")
    out = txtSileORCA(f)
    E = out.read_energy[:]()
    assert len(E) == 2
    assert pytest.approx(E[0].total) == pytest.approx(-3081.2651523095283)
    assert pytest.approx(E[1].total) == pytest.approx(-3081.2651523149702)
    assert pytest.approx(E[1].vdw) == pytest.approx(-0.011180550414138613)
    E = out.read_energy[-1]()
    assert pytest.approx(E.total) == pytest.approx(-3081.2651523149702)
    assert pytest.approx(E.vdw) == pytest.approx(-0.011180550414138613)


def test_read_geometry(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    G = out.read_geometry[:]()
    assert G[0].xyz[0, 0] == pytest.approx(0.421218019838, abs=1e-6)
    assert G[0].xyz[1, 0] == pytest.approx(1.578781980162, abs=1e-6)
    assert G[1].xyz[0, 0] == pytest.approx(0.421218210279, abs=1e-6)
    assert G[1].xyz[1, 0] == pytest.approx(1.578781789721, abs=1e-6)
    G = out.read_geometry[-1]()
    assert G.xyz[0, 0] == pytest.approx(0.421218210279, abs=1e-6)
    assert G.xyz[1, 0] == pytest.approx(1.578781789721, abs=1e-6)
    assert G.xyz[0, 1] == pytest.approx(0.0, abs=1e-6)
    assert G.xyz[1, 1] == pytest.approx(0.0, abs=1e-6)
    assert G.atoms[0].tag == "N"
    assert G.atoms[1].tag == "O"


def test_multiple_calls(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    N = out.read_electrons[:]()
    assert len(N) == 2
    E = out.read_energy[:]()
    assert len(E) == 2
    G = out.read_geometry[:]()
    assert len(G) == 2
    N = out.read_electrons[:]()
    assert len(N) == 2


def test_info_no(sisl_files):
    f = sisl_files("orca", "phenalenyl", "molecule_property.txt")
    out = txtSileORCA(f)
    assert out.info.no == 284


def test_gtensor(sisl_files):
    # file without g-tensor
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    assert out.read_gtensor() is None

    # file with g-tensor
    f = sisl_files("orca", "phenalenyl", "molecule_property.txt")
    out = txtSileORCA(f)
    G = out.read_gtensor()

    assert G.multiplicity == 2
    assert G.tensor[0, 0] == pytest.approx(2.002750)
    assert G.vectors[0, 1] == pytest.approx(-0.009799, abs=1e-4)
    for i in range(3):
        v = G.vectors[i]
        assert v.dot(v) == pytest.approx(1)
    assert G.eigenvalues[0] == pytest.approx(2.002127)


def test_hyperfine_coupling(sisl_files):
    # file without hyperfine_coupling tensors
    f = sisl_files("orca", "nitric_oxide", "molecule_property.txt")
    out = txtSileORCA(f)
    assert out.read_hyperfine_coupling(units="MHz") is None

    # file with hyperfine_coupling tensors
    f = sisl_files("orca", "phenalenyl", "molecule_property.txt")
    out = txtSileORCA(f)
    A = out.read_hyperfine_coupling(units="MHz")
    assert len(A) == 22
    assert A[0].iso == pytest.approx(-23.380794)
    assert A[1].ia == 1
    assert A[1].species == "C"
    assert A[1].isotope == 13
    assert A[1].spin == 0.5
    assert A[1].prefactor == pytest.approx(134.190303)
    assert A[1].tensor[0, 1] == pytest.approx(-0.320129)
    assert A[1].tensor[2, 2] == pytest.approx(68.556557)
    assert A[1].vectors[1, 0] == pytest.approx(0.407884)
    for i in range(3):
        v = A[1].vectors[i]
        assert v.dot(v) == pytest.approx(1)
    assert A[1].eigenvalues[1] == pytest.approx(5.523380)
    assert A[1].iso == pytest.approx(26.247902)
    assert A[12].species == "C"
    assert A[13].species == "H"


def test_hyperfine_coupling_units(sisl_files):
    f = sisl_files("orca", "phenalenyl", "molecule_property.txt")
    out = txtSileORCA(f)
    A = out.read_hyperfine_coupling()
    B = out.read_hyperfine_coupling(units="eV")
    C = out.read_hyperfine_coupling(units={"energy": "eV"})

    assert A[0].iso == B[0].iso == C[0].iso

    A = out.read_hyperfine_coupling(units=("MHz", "Ang"))
    B = out.read_hyperfine_coupling(units={"energy": "MHz"})

    assert A[0].iso == B[0].iso
    assert A[0].iso != C[0].iso



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl.messages as sm
from sisl import Atom, Cuboid, Geometry, Lattice
from sisl._core.sparse_geometry import *
from sisl.geom import fcc, graphene

pytestmark = [
    pytest.mark.sparse,
    pytest.mark.sparse_geometry,
    pytest.mark.sparse_orbital,
]


@pytest.mark.parametrize("n0", [1, 3])
@pytest.mark.parametrize("n1", [1, 4])
@pytest.mark.parametrize("n2", [1, 2])
def test_sparse_orbital_symmetric(n0, n1, n2):
    g = fcc(1.0, Atom(1, R=1.5)) * 2
    s = SparseOrbital(g)
    s.construct([[0.1, 1.51], [1, 2]])
    s = s.tile(n0, 0).tile(n1, 1).tile(n2, 2)
    no = s.geometry.no

    nnz = 0
    for io in range(no):
        # orbitals connecting to io
        edges = s.edges(io)
        # Figure out the transposed supercell indices of the edges
        isc = -s.geometry.o2isc(edges)
        # Convert to supercell
        IO = s.geometry.lattice.sc_index(isc) * no + io
        # Figure out if 'io' is also in the back-edges
        for jo, edge in zip(IO, edges % no):
            assert jo in s.edges(edge)
            nnz += 1

    # Check that we have counted all nnz
    assert s.nnz == nnz


@pytest.mark.parametrize("n0", [1, 3])
@pytest.mark.parametrize("n1", [1, 4])
@pytest.mark.parametrize("n2", [1, 2])
@pytest.mark.parametrize("axis", [0, 1])
def test_sparse_orbital_append(n0, n1, n2, axis):
    g = fcc(1.0, Atom(1, R=1.98)) * 2
    dists = np.insert(g.distance(0, R=g.maxR()) + 0.001, 0, 0.001)
    connect = np.arange(dists.size, dtype=np.float64) / 5
    s = SparseOrbital(g)
    s.construct([dists, connect])
    s = s.tile(2, 0).tile(2, 1).tile(2, 2)
    s1 = s.tile(n0, 0).tile(n1, 1).tile(n2, 2)
    s2 = s1.copy()
    # Resulting full sparse-geometry
    sf = s1.tile(2, axis)
    # Ensure the test works for empty rows
    for i in range(sf.shape[0]):
        sf._csr._extend_empty(i, 11)

    # Now perform some appends and randomizations
    idx1 = np.arange(s1.na)
    idx2 = np.arange(s2.na)

    np.random.seed(42)
    shuffle = np.random.shuffle

    # Test 4 permutations
    for _ in range(3):
        shuffle(idx1)
        shuffle(idx2)
        sout = s1.sub(idx1).append(s2.sub(idx2), axis)
        sout.finalize()

        s = sf.sub(np.concatenate([idx1, s1.na + idx2]))
        assert sout.spsame(s)
        s.finalize()
        assert np.allclose(s._csr._D, sout._csr._D)


@pytest.mark.parametrize("n0", [1, 3])
@pytest.mark.parametrize("n1", [1, 4])
@pytest.mark.parametrize("n2", [1, 2])
@pytest.mark.parametrize("axis", [1, 2])
def test_sparse_orbital_append_scale(n0, n1, n2, axis):
    g = fcc(1.0, Atom(1, R=1.98)) * 2
    dists = np.insert(g.distance(0, R=g.maxR()) + 0.001, 0, 0.001)
    connect = np.arange(dists.size, dtype=np.float64) / 5
    s = SparseOrbital(g)
    s.construct([dists, connect])
    s = s.tile(2, 0).tile(2, 1).tile(2, 2)
    s1 = s.tile(n0, 0).tile(n1, 1).tile(n2, 2)
    s2 = s1.copy()
    # Resulting full sparse-geometry
    sf = s1.tile(2, axis)
    for i in range(sf.shape[0]):
        sf._csr._extend_empty(i, 11)

    # Now perform some appends and randomizations
    idx1 = np.arange(s1.na)
    idx2 = np.arange(s2.na)

    np.random.seed(42)
    shuffle = np.random.shuffle

    # Test 4 permutations
    for _ in range(3):
        shuffle(idx1)
        shuffle(idx2)

        s = sf.sub(np.concatenate([idx1, s1.na + idx2]))
        s.finalize()

        sout = s1.sub(idx1).append(s2.sub(idx2), axis, scale=(2.0, 0))
        sout = (sout + sout.transpose()) * 0.5
        assert sout.spsame(s)
        sout.finalize()
        assert np.allclose(s._csr._D, sout._csr._D)

        sout = s1.sub(idx1).append(s2.sub(idx2), axis, scale=(0.0, 2.0))
        sout.finalize()
        # Ensure that some elements are not the same!
        assert not np.allclose(s._csr._D, sout._csr._D)
        sout = (sout + sout.transpose()) * 0.5
        assert sout.spsame(s)
        sout.finalize()
        assert np.allclose(s._csr._D, sout._csr._D)


def test_sparse_orbital_hermitian():
    g = Geometry([0] * 3, Atom(1, R=1), lattice=Lattice(1, nsc=[3, 1, 1]))

    for f in [True, False]:
        spo = SparseOrbital(g)
        spo[0, 0] = 1.0

        # Create only a single coupling to the neighouring element
        spo[0, 1] = 2.0

        if f:
            spo.finalize()

        assert spo.nnz == 2

        # implicit sort
        spoT = spo.transpose()
        assert spoT.finalized
        assert spoT.nnz == 2
        assert spoT[0, 0] == 1.0
        assert spoT[0, 1] == 0.0
        assert spoT[0, 2] == 2.0

        spoH = (spo + spoT) * 0.5

        assert spoH.nnz == 3
        assert spoH[0, 0] == 1.0
        assert spoH[0, 1] == 1.0
        assert spoH[0, 2] == 1.0


def test_sparse_orbital_sub_orbital():
    a0 = Atom(1, R=(1.1, 1.4, 1.6))
    a1 = Atom(2, R=(1.3, 1.1))
    g = Geometry([[0, 0, 0], [1, 1, 1]], [a0, a1], lattice=Lattice(2, nsc=[3, 1, 1]))
    g = g.tile(3, 0)
    assert g.no == 15

    spo = SparseOrbital(g)
    for io in range(g.no):
        spo[io, io] = io + 1
        spo[io, io + g.no - 1] = io - 2
        spo[io, io + 1] = io + 2
    # Ensure we have a Hermitian matrix
    spo = spo + spo.transpose()

    # orbitals on first atom (a0)
    rem_sub = [(0, [1, 2]), ([0, 2], 1), (2, [0, 1]), (a0[0], [1, 2])]
    for rem, sub in rem_sub:
        spo_rem = spo.remove_orbital(0, rem)
        spo_sub = spo.sub_orbital(0, sub)
        assert spo_rem.spsame(spo_sub)
        atoms = spo_rem.geometry.atoms
        assert atoms == spo_sub.geometry.atoms
        assert atoms.nspecies == 3
        assert (atoms.species == 0).sum() == 2
        assert (atoms.species == 1).sum() == 3
        assert (atoms.species == 2).sum() == 1

        spo_rem = spo.remove_orbital(a0, rem)
        spo_sub = spo.sub_orbital(a0, sub)
        assert spo_rem.spsame(spo_sub)
        atoms = spo_rem.geometry.atoms
        assert atoms == spo_sub.geometry.atoms
        assert atoms.nspecies == 2
        assert (atoms.species == 0).sum() == 3
        assert (atoms.species == 1).sum() == 3

    # orbitals on second atom (a1)
    rem_sub = [
        (0, [1]),
        (a1[0], 1),
        (0, a1[1]),
    ]
    for rem, sub in rem_sub:
        spo_rem = spo.remove_orbital(1, rem)
        spo_sub = spo.sub_orbital(1, sub)
        assert spo_rem.spsame(spo_sub)
        atoms = spo_rem.geometry.atoms
        assert atoms == spo_sub.geometry.atoms
        assert atoms.nspecies == 3
        assert (atoms.species == 0).sum() == 3
        assert (atoms.species == 1).sum() == 2
        assert (atoms.species == 2).sum() == 1

        spo_rem = spo.remove_orbital(a1, rem)
        spo_sub = spo.sub_orbital(a1, sub)
        assert spo_rem.spsame(spo_sub)
        atoms = spo_rem.geometry.atoms
        assert atoms == spo_sub.geometry.atoms
        assert atoms.nspecies == 2
        assert (atoms.species == 0).sum() == 3
        assert (atoms.species == 1).sum() == 3

    spo_rem = spo.remove_orbital([0, 1], 0)
    spo_sub = spo.sub_orbital(0, [1, 2]).sub_orbital(1, 1)
    assert spo_rem.spsame(spo_sub)

    spo_rem = spo.remove_orbital([0, 1], 1)
    spo_sub = spo.sub_orbital(0, [0, 2]).sub_orbital(1, 0)
    assert spo_rem.spsame(spo_sub)


def test_sparse_orbital_sub_orbital_nested():
    """
    Doing nested or multiple subs that exposes the same sub-atom
    should ultimately re-use the existing atom.

    However, due to the renaming of the tags
    """
    a0 = Atom(1, R=(1.1, 1.4, 1.6))
    a1 = Atom(2, R=(1.3, 1.1))
    g = Geometry([[0, 0, 0], [1, 1, 1]], [a0, a1], lattice=Lattice(2, nsc=[3, 1, 1]))
    g = g.repeat(3, 0)
    assert g.no == 15

    spo = SparseOrbital(g)
    for io in range(g.no):
        spo[io, io] = io + 1
        spo[io, io + g.no - 1] = io - 2
        spo[io, io + 1] = io + 2
    # Ensure we have a Hermitian matrix
    spo = spo + spo.transpose()

    assert spo.geometry.atoms.nspecies == 2
    sub0 = spo.remove_orbital(0, 1)
    assert sub0.geometry.atoms.nspecies == 3
    # this will *replace* since we take all atoms of a given species
    sub01 = sub0.remove_orbital(0, 1)
    assert sub01.geometry.atoms.nspecies == 3

    # while this creates the same atom as the above two steps
    # it cannot recognize it as the same atom due
    # to different tag names
    sub = sub01.sub_orbital(1, 0)
    assert sub.geometry.atoms.nspecies == 4


def test_sparse_orbital_replace_simple():
    """
    Replacing parts of a sparse-orbital matrix is quite tricky.

    Here we check a few things with graphene
    """
    gr = graphene()
    spo = SparseOrbital(gr)
    # create the sparse-orbital
    spo.construct([(0.1, 1.45), (0, 2.7)])

    # create 4x4 geoemetry
    spo44 = spo.tile(4, 0).tile(4, 1)

    # now replace every position that can be replaced
    for position in range(0, spo44.geometry.na, 2):
        # replace both atoms
        new = spo44.replace([position, position + 1], spo)
        assert np.fabs((new - spo44)._csr._D).sum() == 0.0

        # replace both atoms (reversed)
        # When swapping it is not the same
        new = spo44.replace([position, position + 1], spo, [1, 0])
        assert np.fabs((new - spo44)._csr._D).sum() != 0.0
        pvt = np.arange(spo44.na)
        pvt[position] = position + 1
        pvt[position + 1] = position
        assert np.unique(pvt).size == pvt.size
        new_pvt = spo44.sub(pvt)
        assert np.fabs((new - new_pvt)._csr._D).sum() == 0.0

        # replace first atom
        new = spo44.replace([position], spo, 0)
        assert np.fabs((new - spo44)._csr._D).sum() == 0.0

        # replace second atom
        new = spo44.replace([position + 1], spo, 1)
        assert np.fabs((new - spo44)._csr._D).sum() == 0.0


def test_sparse_orbital_replace_species_warn():
    """
    Replacing an atom with another one but retaining couplings.
    """
    C = graphene()
    spC = SparseOrbital(C)
    spC.construct([(0.1, 1.45), (0, 2.7)])

    N = graphene(atoms=C.atoms[1].copy(Z=5))
    spN = SparseOrbital(N)
    spN.construct([(0.1, 1.45), (0, 2.7)])

    # Replace
    with pytest.warns(sm.SislWarning):
        sp = spC.copy()
        sp.replace(0, spN, 0)
    with pytest.warns(sm.SislWarning):
        sp = spC.copy()
        sp.replace(1, spN, 1)


def test_sparse_orbital_replace_hole():
    """Create a big graphene flake remove a hole (1 orbital system)"""
    g = graphene(orthogonal=True)
    spo = SparseOrbital(g)
    # create the sparse-orbital
    spo.construct([(0.1, 1.45), (0, 2.7)])

    # create 10x10 geoemetry
    nx, ny = 10, 10
    big = spo.tile(10, 0).tile(10, 1)
    hole = spo.tile(6, 0).tile(6, 1)
    hole = hole.remove(hole.close(hole.center(), R=3))

    def create_sp(geom):
        spo = SparseOrbital(geom)
        # create the sparse-orbital
        spo.construct([(0.1, 1.45), (0, 2.7)])
        return spo

    # now replace every position that can be replaced
    for y in [0, 2, 3]:
        for x in [1, 2, 3]:
            cube = Cuboid(hole.lattice.cell, origin=g.lattice.offset([x, y, 0]) - 0.1)
            atoms = big.within(cube)
            assert len(atoms) == 4 * 6 * 6
            new = big.replace(atoms, hole)
            new_copy = create_sp(new.geometry)
            assert np.fabs((new - new_copy)._csr._D).sum() == 0.0


def test_sparse_orbital_replace_hole_norbs():
    """Create a big graphene flake remove a hole (multiple orbitals)"""
    a1 = Atom(5, R=(1.44, 1.44))
    a2 = Atom(7, R=(1.44, 1.44, 1.44))
    g = graphene(atoms=[a1, a2], orthogonal=True)
    spo = SparseOrbital(g)

    def func(self, ia, atoms, atoms_xyz=None):
        geom = self.geometry

        def a2o(idx):
            return geom.a2o(idx, True)

        io = a2o(ia)
        idx = self.geometry.close(ia, R=[0.1, 1.44], atoms=atoms, atoms_xyz=atoms_xyz)
        idx = list(map(a2o, idx))
        self[io, idx[0]] = 0
        for i in io:
            self[i, idx[1]] = 2.7

    # create the sparse-orbital
    spo.construct(func)

    # create 10x10 geoemetry
    nx, ny = 10, 10
    big = spo.tile(10, 0).tile(10, 1)
    hole = spo.tile(6, 0).tile(6, 1)
    hole = hole.remove(hole.close(hole.center(), R=3))

    def create_sp(geom):
        spo = SparseOrbital(geom)
        # create the sparse-orbital
        spo.construct(func)
        return spo

    # now replace every position that can be replaced
    for y in [0, 3]:
        for x in [1, 3]:
            cube = Cuboid(hole.lattice.cell, origin=g.lattice.offset([x, y, 0]) - 0.1)
            atoms = big.within(cube)
            assert len(atoms) == 4 * 6 * 6
            new = big.replace(atoms, hole)
            new_copy = create_sp(new.geometry)
            assert np.fabs((new - new_copy)._csr._D).sum() == 0.0


def test_translate_sparse_orbitals():
    # Build the geometry
    H = Atom(1, (1, 2, 3))
    graph = graphene(atoms=H)
    assert np.allclose(graph.nsc, [3, 3, 1])

    # Build a dummy matrix with onsite terms and just one coupling term
    matrix = SparseOrbital(graph)
    matrix[0, 0] = 1
    matrix[1, 1] = 2
    matrix[0, 1] = 3
    matrix[0, 4] = 4

    # Translate the second atom
    transl = matrix._translate_atoms_sc([[0, 0, 0], [1, 0, 0]])

    # Check that the new auxiliary cell is correct.
    assert np.allclose(transl.nsc, [3, 1, 1])
    assert np.allclose(transl.shape, [6, 18, 1])

    # Check coordinates
    assert np.allclose(transl.geometry[0], matrix.geometry[0])
    assert np.allclose(transl.geometry[1], matrix.geometry[1] + matrix.geometry.cell[0])

    # Assert that the matrix elements have been translated
    assert transl[0, 0] == 1
    assert transl[1, 1] == 2
    assert transl[0, 1] == 3
    assert transl[0, 6 + 4] == 4

    # Translate back to unit cell
    uc_matrix = transl.translate2uc()

    # Check the auxiliary cell, coordinates and the matrix elements
    assert np.allclose(uc_matrix.nsc, [1, 1, 1])
    assert np.allclose(uc_matrix.shape, [6, 6, 1])

    assert np.allclose(uc_matrix.geometry.xyz, matrix.geometry.xyz)

    assert uc_matrix[0, 0] == 1
    assert uc_matrix[1, 1] == 2
    assert uc_matrix[0, 1] == 3
    assert uc_matrix[0, 4] == 4

    # Instead, test atoms and axes arguments to avoid any translation.
    for kwargs in [{"atoms": [0]}, {"axes": [1, 2]}]:
        not_uc_matrix = transl.translate2uc(**kwargs)

        # Check the auxiliary cell, coordinates and the matrix elements
        assert np.allclose(not_uc_matrix.nsc, transl.nsc)
        assert np.allclose(not_uc_matrix.shape, transl.shape)

        assert np.allclose(not_uc_matrix.geometry.xyz, transl.geometry.xyz)

        assert np.allclose(not_uc_matrix._csr.todense(), transl._csr.todense())

    # Now, translate both atoms
    transl_both = uc_matrix._translate_atoms_sc([[-1, 0, 0], [1, 0, 0]])

    # Check the auxiliary cell, coordinates and the matrix elements
    assert np.allclose(transl_both.nsc, [5, 1, 1])
    assert np.allclose(transl_both.shape, [6, 6 * 5, 1])

    assert np.allclose(
        transl_both.geometry[0], uc_matrix.geometry[0] - uc_matrix.geometry.cell[0]
    )
    assert np.allclose(
        transl_both.geometry[1], uc_matrix.geometry[1] + uc_matrix.geometry.cell[0]
    )

    assert transl_both[0, 0] == 1
    assert transl_both[1, 1] == 2
    assert transl_both[0, 1] == 3
    assert transl_both[0, 6 + 4] == 4



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl.viz.processors.matrix import (
    determine_color_midpoint,
    get_geometry_from_matrix,
    get_matrix_mode,
    get_orbital_sets_positions,
    matrix_as_array,
    sanitize_matrix_arrows,
)

pytestmark = [pytest.mark.viz, pytest.mark.processors]


def test_orbital_positions():

    C = sisl.Atom(
        6,
        orbitals=[
            sisl.AtomicOrbital("2s"),
            sisl.AtomicOrbital("2px"),
            sisl.AtomicOrbital("2py"),
            sisl.AtomicOrbital("2pz"),
            sisl.AtomicOrbital("2px"),
            sisl.AtomicOrbital("2py"),
            sisl.AtomicOrbital("2pz"),
        ],
    )

    H = sisl.Atom(1, orbitals=[sisl.AtomicOrbital("1s")])

    positions = get_orbital_sets_positions([C, H])

    assert len(positions) == 2

    assert positions[0] == [0, 1, 4]
    assert positions[1] == [0]


def test_get_geometry_from_matrix():

    geom = sisl.geom.graphene()

    matrix = sisl.Hamiltonian(geom)

    assert get_geometry_from_matrix(matrix) is geom

    geom_copy = geom.copy()

    assert get_geometry_from_matrix(matrix, geom_copy) is geom_copy

    # Check that if we pass something without an associated geometry
    # but we provide a geometry it will work
    assert get_geometry_from_matrix(np.array([1, 2]), geom) is geom


def test_matrix_as_array():

    matrix = sisl.SparseCSR((2, 2, 2))

    matrix[0, 0, 0] = 1
    matrix[0, 0, 1] = 2

    array = matrix_as_array(matrix, fill_value=0)
    assert np.allclose(array, np.array([[1, 0], [0, 0]]))

    array = matrix_as_array(matrix, dim=1, fill_value=0)
    assert np.allclose(array, np.array([[2, 0], [0, 0]]))

    array = matrix_as_array(matrix)
    assert array[0, 0] == 1
    assert np.isnan(array).sum() == 3

    # Check that it can work with auxiliary supercells
    geom = sisl.geom.graphene(
        atoms=sisl.Atom("C", orbitals=[sisl.AtomicOrbital("2pz")])
    )
    matrix = sisl.Hamiltonian(geom)

    array = matrix_as_array(matrix)
    assert array.shape == matrix.shape[:-1]

    array = matrix_as_array(matrix, isc=1)
    assert array.shape == (geom.no, geom.no)

    # Check that a numpy array is kept untouched
    matrix = np.array([[1, 2], [3, 4]])
    assert np.allclose(matrix_as_array(matrix), matrix)


def test_determine_color_midpoint():

    # With the matrix containing only positive values
    matrix = np.array([1, 2])

    assert determine_color_midpoint(matrix) is None
    assert determine_color_midpoint(matrix, cmid=1, crange=(0, 1)) == 1
    assert determine_color_midpoint(matrix, crange=(0, 1)) is None

    # With the matrix containing only negative values
    matrix = np.array([-1, -2])

    assert determine_color_midpoint(matrix) is None
    assert determine_color_midpoint(matrix, cmid=1, crange=(0, 1)) == 1
    assert determine_color_midpoint(matrix, crange=(0, 1)) is None

    # With the matrix containing both positive and negative values
    matrix = np.array([-1, 1])

    assert determine_color_midpoint(matrix) == 0
    assert determine_color_midpoint(matrix, cmid=1, crange=(-1, 1)) == 1
    assert determine_color_midpoint(matrix, crange=(-1, 1)) is None


def test_get_matrix_mode():

    geom = sisl.geom.graphene()

    matrix = sisl.SparseAtom(geom)
    assert get_matrix_mode(matrix) == "atoms"

    matrix = sisl.Hamiltonian(geom)
    assert get_matrix_mode(matrix) == "orbitals"

    matrix = sisl.SparseCSR((2, 2))
    assert get_matrix_mode(matrix) == "orbitals"

    matrix = np.array([[1, 2], [3, 4]])
    assert get_matrix_mode(matrix) == "orbitals"


def test_sanitize_matrix_arrows():

    arrows = {}
    assert sanitize_matrix_arrows(arrows) == [{"center": "middle"}]

    geom = sisl.geom.graphene()
    data = sisl.Hamiltonian(geom, dim=2)
    data[0, 0, 0] = 1
    data[0, 0, 1] = 2

    arrows = [{"data": data}]
    sanitized = sanitize_matrix_arrows(arrows)

    assert len(sanitized) == 1
    assert sanitized[0]["data"].shape == data.shape
    san_data = sanitized[0]["data"]
    assert san_data[0, 0, 0] == 1
    assert san_data[0, 0, 1] == -2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import scipy as sc

from sisl import Atom, Geometry, Lattice, SislWarning
from sisl._core.sparse_geometry import *
from sisl.geom import fcc, graphene

pytestmark = [pytest.mark.sparse, pytest.mark.sparse_geometry]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            self.g = fcc(1.0, Atom(1, R=1.495)) * 2
            self.s1 = SparseAtom(self.g)
            self.s2 = SparseAtom(self.g, 2)

    return t()


class TestSparseAtom:
    def test_fail_align1(self, setup):
        s = SparseAtom(setup.g * 2)
        str(s)
        with pytest.raises(ValueError):
            setup.s1.spalign(s)

    def test_align1(self, setup):
        s = SparseAtom(setup.g)
        setup.s1.spalign(s)
        setup.s1.spalign(s._csr)

    def test_create1(self, setup):
        setup.s1[0, [1, 2, 3]] = 1
        assert setup.s1.nnz == 3
        assert setup.s1[0, 1] == 1
        setup.s1[2, [1, 2, 3]] = 1
        assert setup.s1.nnz == 6
        setup.s1.empty(keep_nnz=True)
        assert setup.s1.nnz == 6
        setup.s1.empty()
        assert setup.s1.nnz == 0
        # nan produces zeros
        setup.s1[0, 0] = np.nan
        assert setup.s1.nnz == 1
        setup.s1.empty()

    def test_create2(self, setup):
        setup.s1[0, [1, 2, 3], (1, 1, 1)] = 1
        assert setup.s1[0, 2, (1, 1, 1)] == 1
        setup.s1.empty()

    def test_create3(self, setup):
        setup.s2[0, [1, 2, 3], (1, 1, 1)] = [1, 2]
        assert np.allclose(setup.s2[0, 2, (1, 1, 1)], [1, 2])
        assert setup.s2[0, 2, 0, (1, 1, 1)] == 1
        assert setup.s2[0, 2, 1, (1, 1, 1)] == 2
        setup.s2.empty()

    def test_in1(self, setup):
        setup.s1[0, [1, 2, 3]] = 1
        assert setup.s1[0, 2] == 1
        assert [0, 4] not in setup.s1
        assert [0, 2] in setup.s1
        setup.s1.empty()

    def test_reset1(self, setup):
        setup.s1[0, [1, 2, 3], (1, 1, 1)] = 1
        assert setup.s1[0, 2, (1, 1, 1)] == 1
        setup.s1.reset()
        assert setup.s1.nnz == 0
        setup.s1.empty()

    def test_eliminate_zeros1(self, setup):
        setup.s1[0, [1, 2, 3]] = 0
        assert setup.s1.nnz == 3
        setup.s1.eliminate_zeros()
        assert setup.s1.nnz == 0
        setup.s1.empty()

    def test_nonzero1(self, setup):
        s2 = setup.s2.copy()
        s2[0, [0, 2, 3]] = [1, 2]
        s2[3, [1, 2, 3]] = [1, 2]

        r, c = s2.nonzero()
        assert np.allclose(r, [0, 0, 0, 3, 3, 3])
        assert np.allclose(c, [0, 2, 3, 1, 2, 3])
        c = s2.nonzero(only_cols=True)
        assert np.allclose(c, [0, 2, 3, 1, 2, 3])
        r, c = s2.nonzero(atoms=1)
        assert len(r) == 0
        assert len(c) == 0
        r, c = s2.nonzero(atoms=0)
        assert np.allclose(r, [0, 0, 0])
        assert np.allclose(c, [0, 2, 3])
        c = s2.nonzero(atoms=0, only_cols=True)
        assert np.allclose(c, [0, 2, 3])

    def test_create_construct_different_length(self, setup):
        s1 = SparseAtom(setup.g)
        with pytest.raises(ValueError):
            s1.construct([[0.1, 1.5], [1]])

    def test_untile1(self, setup):
        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * 2)
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.untile(2, 2).untile(2, 1).untile(2, 0)
        assert s1.spsame(s2)

        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * [2, 1, 1])
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.untile(2, 0)
        assert s1.spsame(s2)

        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * [1, 2, 1])
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.untile(2, 1)
        assert s1.spsame(s2)

    def test_untile_wrong_usage(self):
        # one should not untile
        geometry = Geometry([0] * 3, Atom(1, R=1.001), Lattice(1, nsc=[1] * 3))
        geometry = geometry.tile(4, 0)
        s = SparseAtom(geometry)
        s.construct([[0.1, 1.01], [1, 2]])
        s[0, 3] = 2
        s[3, 0] = 2

        # check that untiling twice is not the same as untiling 4 times and coupling it
        with pytest.warns(
            SislWarning, match=r"may have connections crossing the entire"
        ):
            s2 = s.untile(2, 0)
        s4 = s.untile(4, 0).tile(2, 0)
        ds = s2 - s4
        ds.finalize()
        assert np.absolute(ds)._csr._D.sum() > 0

    def test_untile_segment_single(self):
        # one should not untile
        geometry = Geometry([0] * 3, Atom(1, R=1.001), Lattice(1, nsc=[1] * 3))
        geometry = geometry.tile(4, 0)
        s = SparseAtom(geometry)
        s.construct([[0.1, 1.01], [1, 2]])
        s[0, 3] = 2
        s[3, 0] = 2

        # check that untiling twice is not the same as untiling 4 times and coupling it
        s4 = s.untile(4, 0).tile(2, 0)
        for seg in range(1, 4):
            sx = s.untile(4, 0, segment=seg).tile(2, 0)
            ds = s4 - sx
            ds.finalize()
            assert np.absolute(ds)._csr._D.sum() == pytest.approx(0.0)

    @pytest.mark.parametrize("axis", [0, 1, 2])
    def test_untile_segment_three(self, axis):
        # one should not untile
        nsc = [3] * 3
        nsc[axis] = 1
        geometry = Geometry([0] * 3, Atom(1, R=1.001), Lattice(1, nsc=nsc))
        geometry = geometry.tile(4, axis)
        s = SparseAtom(geometry)
        s.construct([[0.1, 1.01], [1, 2]])
        s[0, 3] = 2
        s[3, 0] = 2

        # check that untiling twice is not the same as untiling 4 times and coupling it
        s4 = s.untile(4, axis).tile(2, axis)
        for seg in range(1, 4):
            sx = s.untile(4, axis, segment=seg).tile(2, axis)
            ds = s4 - sx
            ds.finalize()
            assert np.absolute(ds)._csr._D.sum() == pytest.approx(0.0)

    def test_unrepeat_setup(self, setup):
        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * ((2, 2, 2), "r"))
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.unrepeat(2, 2).unrepeat(2, 1).unrepeat(2, 0)
        assert s1.spsame(s2)

        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * ([2, 1, 1], "r"))
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.unrepeat(2, 0)
        assert s1.spsame(s2)

        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        s2 = SparseAtom(setup.g * ([1, 2, 1], "r"))
        s2.construct([[0.1, 1.5], [1, 2]])
        s2 = s2.unrepeat(2, 1)
        assert s1.spsame(s2)

    def test_iter(self, setup):
        s1 = SparseAtom(setup.g)
        s1.construct([[0.1, 1.5], [1, 2]])
        i = 0
        for r, c in s1:
            i += 1
        assert i == s1.nnz

    def test_rij_fail1(self, setup):
        s = SparseOrbital(setup.g.copy())
        s.construct([[0.1, 1.5], [1, 2]])
        with pytest.raises(ValueError):
            s.rij(what="none")

    def test_rij_atom(self, setup):
        s = SparseAtom(setup.g.copy())
        s.construct([[0.1, 1.5], [1, 2]])
        atom = s.rij()
        assert atom.spsame(s)

    def test_rij_atom_orbital_compare(self, setup):
        sa = SparseAtom(setup.g.copy())
        so = SparseOrbital(setup.g.copy())
        sa.construct([[0.1, 1.5], [1, 2]])
        so.construct([[0.1, 1.5], [1, 2]])
        atom = sa.rij()
        orbital = so.rij()
        assert atom.spsame(orbital)
        atom = sa.rij()
        orbital = so.rij("atom")
        assert atom.spsame(orbital)
        # This only works because there is 1 orbital per atom
        orbital = so.rij()
        assert so.spsame(orbital)
        so.finalize()
        orbital = so.rij()
        assert so.spsame(orbital)

    def test_sp_orb_remove(self, setup):
        so = SparseOrbital(setup.g.copy())
        so2 = so.remove(0)
        assert so.geometry.na - 1 == so2.geometry.na
        so2 = so.remove([0])
        assert so.geometry.na - 1 == so2.geometry.na
        so2 = so.remove([1])
        assert so.geometry.na - 1 == so2.geometry.na

    def test_sp_orb_remove_atom(self):
        so = SparseOrbital(Geometry([[0] * 3, [1] * 3], [Atom[1], Atom[2]], 2))
        so2 = so.remove(Atom[1])
        assert so.geometry.na - 1 == so2.geometry.na
        assert so.geometry.no - 1 == so2.geometry.no

    def test_remove1(self, setup):
        for i in range(len(setup.g)):
            setup.s1.construct([[0.1, 1.5], [1, 2]])
            s1 = setup.s1.remove(i)
            setup.s1.empty()
            s2 = SparseAtom(setup.g.remove(i))
            s2.construct([[0.1, 1.5], [1, 2]])
            assert s1.spsame(s2)

    def test_sub1(self, setup):
        all = range(len(setup.g))
        for i in range(len(setup.g)):
            setup.s1.construct([[0.1, 1.5], [1, 2]])
            # my new sub
            sub = [j for j in all if i != j]
            s1 = setup.s1.sub(sub)
            setup.s1.empty()
            s2 = SparseAtom(setup.g.sub(sub))
            s2.construct([[0.1, 1.5], [1, 2]])
            assert s1.spsame(s2)
            assert len(s2) == len(sub)

    def test_construct_eta(self, setup):
        s = setup.s1.copy()
        s.construct([[0.1, 1.5], [1, 2]], eta=True)

    def test_tile1(self, setup):
        setup.s1.construct([[0.1, 1.5], [1, 2]])
        setup.s1.finalize()
        s1 = setup.s1.tile(2, 0).tile(2, 1)
        s2 = SparseAtom(setup.g * [2, 2, 1])
        s2.construct([[0.1, 1.5], [1, 2]])
        assert s1.spsame(s2)
        s1.finalize()
        s2.finalize()
        assert np.allclose(s1._csr._D, s2._csr._D)
        s2 = s2.untile(2, 1).untile(2, 0)
        assert setup.s1.spsame(s2)
        s2.finalize()
        assert np.allclose(setup.s1._csr._D, s2._csr._D)
        s1 = s1.untile(2, 1).untile(2, 0)
        assert setup.s1.spsame(s1)
        s1.finalize()
        assert np.allclose(s1._csr._D, setup.s1._csr._D)
        setup.s1.empty()

    def test_tile2(self, setup):
        setup.s1.construct([[0.1, 1.5], [1, 2]])
        setup.s1.finalize()
        s1 = setup.s1.tile(2, 0).tile(2, 1)
        s2 = SparseAtom(setup.g * [2, 2, 1])
        s2.construct([[0.1, 1.5], [1, 2]])
        assert s1.spsame(s2)
        s1.finalize()
        s2.finalize()
        assert np.allclose(s1._csr._D, s2._csr._D)
        s2 = s2.untile(2, 1).untile(2, 0)
        assert setup.s1.spsame(s2)
        s2.finalize()
        assert np.allclose(setup.s1._csr._D, s2._csr._D)
        s1 = s1.untile(2, 1).untile(2, 0)
        assert setup.s1.spsame(s1)
        s1.finalize()
        assert np.allclose(s1._csr._D, setup.s1._csr._D)
        setup.s1.empty()

    def test_repeat1(self, setup):
        setup.s1.construct([[0.1, 1.5], [1, 2]])
        s1 = setup.s1.repeat(2, 0).repeat(2, 1)
        setup.s1.empty()
        s2 = SparseAtom(setup.g * ([2, 2, 1], "r"))
        s2.construct([[0.1, 1.5], [1, 2]])
        assert s1.spsame(s2)
        s1.finalize()
        s2.finalize()
        assert np.allclose(s1._csr._D, s2._csr._D)

    def test_repeat2(self, setup):
        setup.s1.construct([[0.1, 1.5], [1, 2]])
        setup.s1.finalize()
        s1 = setup.s1.repeat(2, 0).repeat(2, 1)
        setup.s1.empty()
        s2 = SparseAtom(setup.g * ([2, 2, 1], "r"))
        s2.construct([[0.1, 1.5], [1, 2]])
        assert s1.spsame(s2)
        s1.finalize()
        s2.finalize()
        assert np.allclose(s1._csr._D, s2._csr._D)

    def test_supercell_poisition1(self, setup):
        g1 = setup.g.copy()
        g2 = setup.g.translate([100, 100, 100])
        # Just check that the atomic coordinates are not equivalent
        # up to 1 angstrom
        assert not np.allclose(g1.xyz, g2.xyz, atol=1)
        s1 = SparseAtom(g1)
        s2 = SparseAtom(g2)
        s1.construct([[0.1, 1.5], [1, 2]])
        s1.finalize()
        s2.construct([[0.1, 1.5], [1, 2]])
        s2.finalize()
        assert s1.spsame(s2)

    def test_set_nsc1(self, setup):
        g = fcc(1.0, Atom(1, R=3.5))
        s = SparseAtom(g)
        s.construct([[0.1, 1.5, 3.5], [1, 2, 3]])
        s.finalize()
        assert s.nnz > 1
        s.set_nsc([1, 1, 1])
        assert s.nnz == 1
        assert s[0, 0] == 1

    def test_set_nsc2(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        s = SparseAtom(g)
        s.construct([[0.1, 1.43], [1, 2]])
        s.finalize()
        assert s.nnz == 8
        s.set_nsc(a=1)
        assert s.nnz == 6
        s.set_nsc([None, 1, 1])
        assert s.nnz == 4
        assert s[0, 0] == 1

    def test_set_nsc3(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        s = SparseAtom(g)

        s.set_nsc((3, 3, 1))
        s.construct([[0.1, 1.43], [1, 2]])

        s55 = s.copy()
        s55.set_nsc((5, 5, 1))
        s77 = s55.copy()
        s77.set_nsc((7, 7, 1))
        s59 = s77.copy()
        s59.set_nsc((5, 9, 1))
        assert s.nnz == s55.nnz
        assert s.nnz == s77.nnz
        assert s.nnz == s59.nnz

        s55.set_nsc((3, 3, 1))
        s77.set_nsc((3, 3, 1))
        s59.set_nsc((3, 3, 1))
        assert s.nnz == s55.nnz
        assert s.nnz == s77.nnz
        assert s.nnz == s59.nnz

    def test_edges1(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        s = SparseAtom(g)
        s.construct([[0.1, 1.43], [1, 2]])
        assert len(s.edges(0)) == 4
        assert len(s.edges(0, exclude=[0])) == 3

    def test_op_numpy_iscalar(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        S = SparseAtom(g, dtype=np.complex128)
        I = np.float32(1)
        # Create initial stuff
        for i in range(10):
            j = range(i, i * 2)
            S[0, j] = i
        S.finalize()

        Ssum = S._csr._D.sum()

        S += I
        assert isinstance(S, SparseAtom)
        assert S._csr._D.sum() == Ssum + S.nnz

        S -= I
        assert isinstance(S, SparseAtom)
        assert S._csr._D.sum() == Ssum

        S *= I
        assert isinstance(S, SparseAtom)
        assert S._csr._D.sum() == Ssum

        S /= I
        assert isinstance(S, SparseAtom)
        assert S._csr._D.sum() == Ssum

        S **= I
        assert isinstance(S, SparseAtom)
        assert S._csr._D.sum() == Ssum

    def test_op_numpy_scalar(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        S = SparseAtom(g)
        I = np.ones(1, dtype=np.complex128)[0]
        # Create initial stuff
        for i in range(10):
            j = range(i, i * 2)
            S[0, j] = i
        S.finalize()

        Ssum = S._csr._D.sum()

        s = S + I
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum + S.nnz

        s = S - I
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum - S.nnz

        s = I + S
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum + S.nnz

        s = S * I
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum

        s = I * S
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum

        s = S / I
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum

        s = S**I
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128
        assert s._csr._D.sum() == Ssum

        s = I**S
        assert isinstance(s, SparseAtom)
        assert s.dtype == np.complex128

    def test_numpy_reduction(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        S = SparseAtom(g)
        I = np.ones(1, dtype=np.complex128)[0]
        # Create initial stuff
        for i in range(2):
            j = range(i, i + 2)
            S[i, j] = 1
        S.finalize()
        assert np.sum(S, axis=(0, 1)) == pytest.approx(1 * 2 * 2)

    def test_sanitize_atoms_assign(self, setup):
        g = graphene(atoms=Atom(6, R=1.43))
        S = SparseAtom(g)
        for i in range(2):
            S[i, 1:4] = 1

    def test_fromsp1(self, setup):
        g = setup.g.repeat(2, 0).tile(2, 1)
        lil = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil[0, [1, 2, 3]] = 1
        lil[1, [2, 4, 1]] = 2
        s1 = SparseAtom.fromsp(g, [lil], unknown_key="hello")
        assert s1.nnz == 6
        assert np.allclose(s1.shape, [g.na, g.na_s, 1])

        assert np.allclose(s1[0, [1, 2, 3]], np.ones([3], np.int32))
        assert np.allclose(s1[1, [1, 2, 4]], np.ones([3], np.int32) * 2)

        # Different instantiating
        s2 = SparseAtom.fromsp(g, lil)
        assert s1.spsame(s2)

    def test_fromsp2(self, setup):
        g = setup.g.repeat(2, 0).tile(2, 1)
        lil1 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil2 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil1[0, [1, 2, 3]] = 1
        lil2[1, [2, 4, 1]] = 2
        s1 = SparseAtom.fromsp(g, [lil1, lil2])
        assert s1.nnz == 6
        assert np.allclose(s1.shape, [g.na, g.na_s, 2])

        assert np.allclose(s1[0, [1, 2, 3], 0], np.ones([3], np.int32))
        assert np.allclose(s1[0, [1, 2, 3], 1], np.zeros([3], np.int32))
        assert np.allclose(s1[1, [1, 2, 4], 0], np.zeros([3], np.int32))
        assert np.allclose(s1[1, [1, 2, 4], 1], np.ones([3], np.int32) * 2)

    def test_fromsp4(self, setup):
        g = setup.g.repeat(2, 0).tile(2, 1)
        lil1 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil2 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil1[0, [1, 2, 3]] = 1
        lil2[1, [2, 4, 1]] = 2

        # Ensure that one does not mix everything.
        with pytest.raises(ValueError):
            SparseAtom.fromsp(setup.g.copy(), [lil1, lil2])

    def test_pickle(self, setup):
        import pickle as p

        g = setup.g.repeat(2, 0).tile(2, 1)
        lil1 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil2 = sc.sparse.lil_matrix((g.na, g.na_s), dtype=np.int32)
        lil1[0, [1, 2, 3]] = 1
        lil2[1, [2, 4, 1]] = 2
        S = SparseAtom.fromsp(g, [lil1, lil2])
        n = p.dumps(S)
        s = p.loads(n)
        assert s.spsame(S)


@pytest.mark.parametrize("n0", [1, 2])
@pytest.mark.parametrize("n1", [1, 3])
@pytest.mark.parametrize("n2", [1, 4])
def test_sparse_atom_symmetric(n0, n1, n2):
    g = fcc(1.0, Atom(1, R=1.5)) * 2
    s = SparseAtom(g)
    s.construct([[0.1, 1.51], [1, 2]])
    s = s.tile(n0, 0).tile(n1, 1).tile(n2, 2)
    na = s.geometry.na

    nnz = 0
    for ia in range(na):
        # orbitals connecting to ia
        edges = s.edges(ia)
        # Figure out the transposed supercell indices of the edges
        isc = -s.geometry.a2isc(edges)
        # Convert to supercell
        IA = s.geometry.lattice.sc_index(isc) * na + ia
        # Figure out if 'ia' is also in the back-edges
        for ja, edge in zip(IA, edges % na):
            assert ja in s.edges(edge)
            nnz += 1

    # Check that we have counted all nnz
    assert s.nnz == nnz


@pytest.mark.parametrize("i", [0, 1, 2])
def test_sparse_atom_transpose_single(i):
    """This is problematic when the sparsity pattern is not *filled*"""
    g = fcc(1.0, Atom(1, R=1.5)) * 3
    s = SparseAtom(g)
    s[i, 2] = 1.0
    s[i, 0] = 2.0
    t = s.transpose()

    assert t.nnz == s.nnz
    assert t[2, i] == pytest.approx(1.0)
    assert t[0, i] == pytest.approx(2.0)


@pytest.mark.parametrize("i", [0, 1, 2])
def test_sparse_atom_transpose_more(i):
    """This is problematic when the sparsity pattern is not *filled*"""
    g = fcc(1.0, Atom(1, R=1.5)) * 3
    s = SparseAtom(g)
    s[i, 2] = 1.0
    s[i, 0] = 2.0
    s[i + 2, 3] = 1.0
    s[i + 2, 5] = 2.0
    t = s.transpose()

    assert t.nnz == s.nnz
    assert t[2, i] == pytest.approx(1.0)
    assert t[0, i] == pytest.approx(2.0)
    assert t[3, i + 2] == pytest.approx(1.0)
    assert t[5, i + 2] == pytest.approx(2.0)


@pytest.mark.parametrize("i", [0, 1, 2])
def test_sparse_orbital_transpose_single(i):
    g = fcc(1.0, Atom(1, R=(1.5, 2.1))) * 3
    s = SparseOrbital(g)
    s[i, 2] = 1.0
    s[i, 0] = 2.0
    t = s.transpose()

    assert t.nnz == s.nnz
    assert t[2, i] == pytest.approx(1.0)
    assert t[0, i] == pytest.approx(2.0)


@pytest.mark.parametrize("i", [0, 1, 2])
def test_sparse_orbital_transpose_more(i):
    g = fcc(1.0, Atom(1, R=(1.5, 2.1))) * 3
    s = SparseOrbital(g)
    s[i, 2] = 1.0
    s[i, 0] = 2.0
    s[i + 3, 4] = 1.0
    s[i + 2, 4] = 2.0
    t = s.transpose()

    assert t.nnz == s.nnz
    assert t[2, i] == pytest.approx(1.0)
    assert t[0, i] == pytest.approx(2.0)
    assert t[4, i + 3] == pytest.approx(1.0)
    assert t[4, i + 2] == pytest.approx(2.0)


def test_sparse_orbital_add_axis(setup):
    g = setup.g.copy()
    s = SparseOrbital(g)
    s.construct([[0.1, 1.5], [1, 2]])
    s1 = s.add(s, axis=2)
    with pytest.warns(SislWarning, match=r"with 0 length"):
        s2 = SparseOrbital(g.append(Lattice([0, 0, 10]), 2).add(g, offset=[0, 0, 5]))
    s2.construct([[0.1, 1.5], [1, 2]])
    assert s1.spsame(s2)


def test_sparse_orbital_add_no_axis():
    from sisl.geom import sc

    with pytest.warns(SislWarning, match=r"with 0 length"):
        g = (sc(1.0, Atom(1, R=1.5)) * 2).add(Lattice([0, 0, 5]))
    s = SparseOrbital(g)
    s.construct([[0.1, 1.5], [1, 2]])
    s1 = s.add(s, offset=[0, 0, 3])
    s2 = SparseOrbital(g.add(g, offset=[0, 0, 3]))
    s2.construct([[0.1, 1.5], [1, 2]])
    assert s1.spsame(s2)


def test_sparse_orbital_sub_orbital():
    atom = Atom(1, (1, 2, 3))
    g = fcc(1.0, atom) * 2
    s = SparseOrbital(g)

    # take out some orbitals
    s1 = s.sub_orbital(atom, 1)
    assert s1.geometry.no == s1.geometry.na

    s2 = s.sub_orbital(atom, atom.orbitals[1])
    assert s1 == s2

    s2 = s.sub_orbital(atom, [atom.orbitals[1]])
    assert s1 == s2

    s2 = s.sub_orbital(atom, [atom.orbitals[1], atom.orbitals[0]])
    assert s2.geometry.atoms[0].orbitals[0] == atom.orbitals[0]
    assert s2.geometry.atoms[0].orbitals[1] == atom.orbitals[1]


def test_translate_sparse_atoms():
    # Build the geometry
    H = Atom(1, (1, 2, 3))
    graph = graphene(atoms=H)
    assert np.allclose(graph.nsc, [3, 3, 1])

    # Build a dummy matrix with onsite terms and just one coupling term
    matrix = SparseAtom(graph)
    matrix[0, 0] = 1
    matrix[1, 1] = 2
    matrix[0, 1] = 3

    # Translate the second atom
    transl = matrix._translate_atoms_sc([[0, 0, 0], [1, 0, 0]])

    # Check that the new auxiliary cell is correct.
    assert np.allclose(transl.nsc, [3, 1, 1])
    assert np.allclose(transl.shape, [2, 6, 1])

    # Check coordinates
    assert np.allclose(transl.geometry[0], matrix.geometry[0])
    assert np.allclose(transl.geometry[1], matrix.geometry[1] + matrix.geometry.cell[0])

    # Assert that the matrix elements have been translated
    assert transl[0, 0] == 1
    assert transl[1, 1] == 2
    assert transl[0, 1] == 0
    assert transl[0, 3] == 3

    # Translate back to unit cell
    uc_matrix = transl.translate2uc()

    # Check the auxiliary cell, coordinates and the matrix elements
    assert np.allclose(uc_matrix.nsc, [1, 1, 1])
    assert np.allclose(uc_matrix.shape, [2, 2, 1])

    assert np.allclose(uc_matrix.geometry.xyz, matrix.geometry.xyz)

    assert uc_matrix[0, 0] == 1
    assert uc_matrix[1, 1] == 2
    assert uc_matrix[0, 1] == 3
    assert uc_matrix[0, 3] == 0

    # Instead, test atoms and axes arguments to avoid any translation.
    for kwargs in [{"atoms": [0]}, {"axes": [1, 2]}]:
        not_uc_matrix = transl.translate2uc(**kwargs)

        # Check the auxiliary cell, coordinates and the matrix elements
        assert np.allclose(not_uc_matrix.nsc, transl.nsc)
        assert np.allclose(not_uc_matrix.shape, transl.shape)

        assert np.allclose(not_uc_matrix.geometry.xyz, transl.geometry.xyz)

        assert np.allclose(not_uc_matrix._csr.todense(), transl._csr.todense())

    # Now, translate both atoms
    transl_both = uc_matrix._translate_atoms_sc([[-1, 0, 0], [1, 0, 0]])

    # Check the auxiliary cell, coordinates and the matrix elements
    assert np.allclose(transl_both.nsc, [5, 1, 1])
    assert np.allclose(transl_both.shape, [2, 10, 1])

    assert np.allclose(
        transl_both.geometry[0], uc_matrix.geometry[0] - uc_matrix.geometry.cell[0]
    )
    assert np.allclose(
        transl_both.geometry[1], uc_matrix.geometry[1] + uc_matrix.geometry.cell[0]
    )

    assert transl_both[0, 0] == 1
    assert transl_both[1, 1] == 2
    assert transl_both[0, 1] == 0
    assert transl_both[0, 3] == 3


def test_sanitize_orbs_assign():
    g = graphene(atoms=Atom(6, R=[1.43, 1.66]))
    S = SparseOrbital(g)
    for i in range(2):
        S[i, 1:4] = 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry
from sisl.io.vasp.car import *

pytestmark = [pytest.mark.io, pytest.mark.vasp]


def test_geometry_car_mixed(sisl_tmp):
    f = sisl_tmp("test_read_write.POSCAR")

    atoms = [Atom[1], Atom[2], Atom[2], Atom[1], Atom[1], Atom[2], Atom[3]]
    xyz = np.random.rand(len(atoms), 3)
    geom = Geometry(xyz, atoms, 100)

    geom.write(carSileVASP(f, "w"))

    assert carSileVASP(f).read_geometry() == geom


def test_geometry_car_group(sisl_tmp):
    f = sisl_tmp("test_sort.POSCAR")

    atoms = [Atom[1], Atom[2], Atom[2], Atom[1], Atom[1], Atom[2], Atom[3]]
    xyz = np.random.rand(len(atoms), 3)
    geom = Geometry(xyz, atoms, 100)

    geom.write(carSileVASP(f, "w"), group_species=True)

    assert carSileVASP(f).read_geometry() != geom
    geom = carSileVASP(f).geometry_group(geom)
    assert carSileVASP(f).read_geometry() == geom


def test_geometry_car_allsame(sisl_tmp):
    f = sisl_tmp("test_read_write.POSCAR")

    atoms = Atom[1]
    xyz = np.random.rand(10, 3)
    geom = Geometry(xyz, atoms, 100)

    geom.write(carSileVASP(f, "w"))

    assert carSileVASP(f).read_geometry() == geom


def test_geometry_car_dynamic(sisl_tmp):
    f = sisl_tmp("test_dynamic.POSCAR")

    atoms = Atom[1]
    xyz = np.random.rand(10, 3)
    geom = Geometry(xyz, atoms, 100)

    read = carSileVASP(f)

    # no dynamic (direct geometry)
    geom.write(carSileVASP(f, "w"), dynamic=None)
    g, dyn = read.read_geometry(ret_dynamic=True)
    assert dyn is None

    geom.write(carSileVASP(f, "w"), dynamic=False)
    g, dyn = read.read_geometry(ret_dynamic=True)
    assert not np.any(dyn)

    geom.write(carSileVASP(f, "w"), dynamic=True)
    g, dyn = read.read_geometry(ret_dynamic=True)
    assert np.all(dyn)

    dynamic = [False] * len(geom)
    dynamic[0] = [True, False, True]
    geom.write(carSileVASP(f, "w"), dynamic=dynamic)
    g, dyn = read.read_geometry(ret_dynamic=True)
    assert np.array_equal(dynamic[0], dyn[0])
    assert not np.any(dyn[1:])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from numbers import Integral, Real

import numpy as np
import pytest

from sisl import Atom, Geometry, Lattice


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )
            C = Atom(Z=6, R=bond * 1.01)
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )

            C = Atom(Z=6, R=[bond * 1.01] * 2)
            self.g2 = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )

    return t()


@pytest.mark.geometry
class TestGeometryReturn:
    def test_fl_o(self, setup):
        # first o is always one element longer than the number of atoms
        assert np.all(setup.g.firsto == [0, 1, 2])
        assert np.all(setup.g.lasto == [0, 1])

        # first o is always one element longer than the number of atoms
        assert np.all(setup.g2.firsto == [0, 2, 4])
        assert np.all(setup.g2.lasto == [1, 3])

    def test_rij1(self, setup):
        d = setup.g.rij(0, 1)
        assert isinstance(d, Real)
        d = setup.g.rij(0, [0, 1])
        assert len(d) == 2
        d = setup.g.rij([0], [0, 1])
        assert len(d) == 1

        # also test orij
        d = setup.g.orij(0, 1)
        assert isinstance(d, Real)
        d = setup.g.orij(0, [0, 1])
        assert len(d) == 2
        d = setup.g.orij([0], [0, 1])
        assert len(d) == 1

    def test_rij2(self, setup):
        d = setup.g.rij([0, 1], [0, 1])
        assert np.allclose(d, [0.0, 0.0])

    def test_osc2uc(self, setup):
        # single value
        d = setup.g.osc2uc(0)
        assert d == 0
        # more values
        d = setup.g.osc2uc([0, 1, 2])
        assert len(d) == 3
        assert np.allclose(d, [0, 1, 0])

    def test_slice1(self, setup):
        d = setup.g[1]
        assert len(d) == 3
        d = setup.g[[1, 2]]
        assert d.shape == (2, 3)
        d = setup.g[1:3]
        assert d.shape == (2, 3)
        d = setup.g[2:4]
        assert d.shape == (2, 3)

        d = setup.g[1, 2]
        assert d == 0.0
        d = setup.g[1, 1:3]
        assert d.shape == (2,)
        d = setup.g[1, :]
        assert d.shape == (3,)
        d = setup.g[2:4, :]
        assert d.shape == (2, 3)
        d = setup.g[2:4, 1]
        assert d.shape == (2,)
        d = setup.g[2:4, 1:3]
        assert d.shape == (2, 2)
        d = setup.g[2:10:2, 1:3]
        assert d.shape == (4, 2)
        d = setup.g[2:10:2, 2]
        assert d.shape == (4,)

        d = setup.g[None, 2]
        assert d.shape == (len(setup.g),)

    def test_a2o(self, setup):
        d = setup.g.a2o(1)
        assert isinstance(d, Integral)
        d = setup.g.a2o([0, 1])
        assert len(d) == 2
        d = setup.g2.a2o(1)
        assert isinstance(d, Integral)
        d = setup.g2.a2o([0, 1])
        assert len(d) == 2

        d = setup.g.a2o(1, True)
        assert len(d) == 1
        d = setup.g.a2o([0, 1], True)
        assert len(d) == 2
        d = setup.g2.a2o(1, True)
        assert len(d) == 2
        d = setup.g2.a2o([0, 1], True)
        assert len(d) == 4

    def test_o2a(self, setup):
        d = setup.g.o2a(1)
        assert isinstance(d, Integral)
        d = setup.g.o2a([0, 1])
        assert len(d) == 2

    def test_axyz(self, setup):
        d = setup.g.axyz(1)
        assert len(d) == 3
        d = setup.g.axyz([0, 1])
        assert d.shape == (2, 3)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.table import *

pytestmark = [pytest.mark.io, pytest.mark.generic]


def test_tbl1(sisl_tmp):
    dat0 = np.arange(2)
    dat1 = np.arange(2) + 1

    io0 = tableSile(sisl_tmp("t0.dat"), "w")
    io1 = tableSile(sisl_tmp("t1.dat"), "w")
    io0.write_data(dat0, dat1)
    io1.write_data((dat0, dat1))

    F0 = io0.file.open().readlines()
    F1 = io1.file.open().readlines()
    assert all([l0 == l1 for l0, l1 in zip(F0, F1)])


def test_tbl2(sisl_tmp):
    dat0 = np.arange(8).reshape(2, 2, 2)
    dat1 = np.arange(8).reshape(2, 2, 2) + 1

    io0 = tableSile(sisl_tmp("t0.dat"), "w")
    io1 = tableSile(sisl_tmp("t1.dat"), "w")
    io0.write_data(dat0, dat1)
    io1.write_data((dat0, dat1))

    F0 = io0.file.open().readlines()
    F1 = io1.file.open().readlines()
    assert all([l0 == l1 for l0, l1 in zip(F0, F1)])


def test_tbl3(sisl_tmp):
    dat0 = np.arange(8).reshape(2, 2, 2)
    dat1 = np.arange(8).reshape(2, 2, 2) + 1
    DAT = np.stack([dat0, dat1])
    DAT.shape = (-1, 2, 2)

    io = tableSile(sisl_tmp("t.dat"), "w")
    io.write_data(dat0, dat1)
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)
    io = tableSile(io.file, "w")
    io.write_data((dat0, dat1))
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)


def test_tbl4(sisl_tmp):
    dat0 = np.arange(8)
    dat1 = np.arange(8) + 1
    DAT = np.stack([dat0, dat1])

    io = tableSile(sisl_tmp("t.dat"), "w")
    io.write_data(dat0, dat1)
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)
    io = tableSile(io.file, "w")
    io.write_data((dat0, dat1))
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)


def test_tbl_automatic_stack(sisl_tmp):
    dat0 = np.arange(4)
    dat1 = np.arange(8).reshape(2, 4) + 1
    DAT = np.vstack([dat0, dat1])

    io = tableSile(sisl_tmp("t.dat"), "w")
    io.write_data(dat0, dat1)
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)
    io = tableSile(io.file, "w")
    io.write_data((dat0, dat1))
    dat = tableSile(io.file, "r").read_data()
    assert np.allclose(dat, DAT)


def test_tbl_accumulate(sisl_tmp):
    DAT = np.arange(12).reshape(3, 4) + 1

    io = tableSile(sisl_tmp("t.dat"), "w")
    with io:
        io.write_data(header="Hello")
        for d in DAT.T:
            # Write a row at a time (otherwise 1D data will be written as a single column)
            io.write_data(d.reshape(-1, 1))
    dat, header = tableSile(io.file, "r").read_data(ret_header=True)
    assert np.allclose(dat, DAT)
    assert header.index("Hello") >= 0


def test_tbl_accumulate_1d(sisl_tmp):
    DAT = np.arange(12).reshape(3, 4) + 1

    io = tableSile(sisl_tmp("t.dat"), "w")
    with io:
        io.write_data(header="Hello")
        for d in DAT:
            # 1D data will be written as a single column
            io.write_data(d)
    dat, header = tableSile(io.file, "r").read_data(ret_header=True)
    assert dat.ndim == 1
    assert np.allclose(dat, DAT.ravel())
    assert header.index("Hello") >= 0


@pytest.mark.parametrize("delimiter", ["\t", " ", ",", ":", "M"])
def test_tbl5(sisl_tmp, delimiter):
    dat0 = np.arange(8)
    dat1 = np.arange(8) + 1
    DAT = np.stack([dat0, dat1])

    io = tableSile(sisl_tmp("t.dat"), "w")
    io.write_data(dat0, dat1, delimiter=delimiter)
    if delimiter in ["\t", " ", ","]:
        dat = tableSile(io.file, "r").read_data()
        assert np.allclose(dat, DAT)
    dat = tableSile(io.file, "r").read_data(delimiter=delimiter)
    assert np.allclose(dat, DAT)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest
import scipy.linalg as sl

from sisl.linalg import eig, eig_destroy, eigh, eigh_destroy

pytestmark = [pytest.mark.linalg, pytest.mark.eig]


def test_eig1():
    np.random.seed(138012)
    a = np.random.rand(10, 10)
    ac = a.copy()
    b = np.random.rand(10, 10)
    bc = b.copy()
    xs, vs = sl.eig(a, b)
    x, v = eig(a, b)
    assert np.allclose(xs, x)
    assert np.allclose(vs, v)
    assert np.allclose(a, ac)
    assert np.allclose(b, bc)


def test_eigh1():
    np.random.seed(1204982)
    a = np.random.rand(10, 10)
    # Symmetrize
    a = a + a.T
    ac = a.copy()
    xs, vs = sl.eigh(a)
    x, v = eigh(a)
    assert np.allclose(xs, x)
    assert np.allclose(vs, v)
    assert np.allclose(a, ac)


def test_eig_d1():
    np.random.seed(138012)
    a = np.random.rand(10, 10)
    b = np.random.rand(10, 10)
    xs, vs = sl.eig(a, b)
    x, v = eig_destroy(a, b)
    assert np.allclose(xs, x)
    assert np.allclose(vs, v)


def test_eigh_d1():
    np.random.seed(1204982)
    a = np.random.rand(10, 10)
    # Symmetrize
    a = a + a.T
    xs, vs = sl.eigh(a)
    x, v = eigh_destroy(a)
    assert np.allclose(xs, x)
    assert np.allclose(vs, v)
